// Anipulator.cpp : Defines the entry point for the application.
//

#define _WIN32_WINNT 0x0400     // WM_MOUSEWHEEL
#include <windows.h>


#include <math.h>


const char* szTitle   = "Julia";
const char* szCaption = "Julia Set Interactive (F1 Help)";

const char* szAbout   = " 2006 bntr.planet.ee\n\nF11\t\tFullScreen\nMouse Wheel\tZoom\n+,-\t\tIterations\nP,<-,->\t\tPolyline\nLeft Mouse\tShift\nM\t\tMandelbrot";
#define     IDM_ABOUT   0x0010




// Window
long        nMouseX;
long        nMouseY;
long        nMouseShiftX;
long        nMouseShiftY;

BYTE        bFullScreen = 0;


long        nW;
long        nH;

#define     ONEBITS2    12
#define     ONEBITS     (ONEBITS2<<1)
#define     ONEBITS21   (ONEBITS2-1)


//
int         nScale = 0;     // ~ -8..8
int         nIterBits = 5;  // 1..6


// Screen
HBITMAP     hbScreen = NULL;

#define     MAXPEN        0x7F
HPEN        hPens[MAXPEN+1];


// Image
HBITMAP     hbBg    = NULL;

HBITMAP     hbBgDib = NULL;
LPBYTE      pBgDibBits     = NULL;  // Mandelbrot x Julia  (Bits for HBITMAP Section)
LPBYTE      pBgDibBitsMand = NULL;  // Mandelbrot          (Only Buffer)
int         nBgDibHeight;
int         nBgDibWidthBytes;

#define     REDRAW_JULIA    0x01
#define     REDRAW_MAND     0x02
#define     REDRAW_ALL      0x03
BYTE        bRedraw = REDRAW_ALL;

BYTE        bMandShow = 1;

BYTE        bHelpShow = 0;

// Polyline
BYTE        bPolylineShow = 1;
#define     MAXPOLYLINE   0x200
int         nPolylineBits = 6;    // Iterations



//========================================== Image

void RedrawBg()
{
    int i,j;

    long nW2 = nW>>1;
    long nH2 = nH>>1;

    long nScaleMult = long( (1 << ONEBITS-7) / pow( 2, nScale/8.0 ) );  // 8scrolls for scale=x2
                                                                        // 7 - default scale
    long niCX;
    long niCY;

    LPBYTE pLine;
    LPBYTE pLineMand;

    long niX,niY,niXX,niYY;
    int  nIter;


    // ------------------- With Mandelbrot Mixing

    if( bMandShow )
    {
        // Redraw Mandelbrot Bits

        if( REDRAW_MAND & bRedraw  &&  bMandShow )
        {
            for( i=0; i<nH; i++ )  {
                pLineMand = pBgDibBitsMand + nW * i;

                for( j=0; j<nW; j++ )  {

                    niCX = (j-nW2) * nScaleMult;
                    niCY = (i-nH2) * nScaleMult;

                    niX = (j-nW2) * nScaleMult;
                    niY = (i-nH2) * nScaleMult;

                    nIter = 0;
                    while( nIter ^ 1<<nIterBits )
                    {    
                        niXX = niX>>ONEBITS2;  niXX *= niXX;
                        niYY = niY>>ONEBITS2;  niYY *= niYY;

                        if( niXX+niYY > 4<<ONEBITS )
                            break;

                        niY = (niX>>ONEBITS2)*(niY>>ONEBITS21) + niCY;
                        niX = niXX - niYY                      + niCX;

                        nIter++;
                    }

                   *pLineMand++ = 255 - (nIter << (7-nIterBits));
                }
            }
        
        }

        // Redraw Julia Bits

        niCX = (nMouseX-nW2) * nScaleMult;
        niCY = (nMouseY-nH2) * nScaleMult;

        for( i=0; i<nH; i++ )
        {
            pLine     = pBgDibBits + nBgDibWidthBytes * (nBgDibHeight-1 - i);
            pLineMand = pBgDibBitsMand + nW * i;

            for( j=0; j<nW; j++ )  {

                niX = (j-nW2) * nScaleMult;
                niY = (i-nH2) * nScaleMult;

                nIter = 0;
                while( nIter ^ 1<<nIterBits )
                {    
                    niXX = niX>>ONEBITS2;  niXX *= niXX;
                    niYY = niY>>ONEBITS2;  niYY *= niYY;

                    if( niXX+niYY > 4<<ONEBITS )
                        break;

                    niY = (niX>>ONEBITS2)*(niY>>ONEBITS21) + niCY;
                    niX = niXX - niYY                      + niCX;

                    nIter++;
                }

                BYTE col = 255 - (nIter << (7-nIterBits));



               *pLine++ = *pLineMand;
               *pLine++ = (col >> 1) + (*pLineMand >> 1);
               *pLine++ = col;

                pLineMand ++;
            }
        }
    }

    // ------------------- Without Mandelbrot
    else
    {
        niCX = (nMouseX-nW2) * nScaleMult;
        niCY = (nMouseY-nH2) * nScaleMult;

        for( i=0; i<nH; i++ )
        {
            pLine = pBgDibBits + nBgDibWidthBytes * (nBgDibHeight-1 - i);

            for( j=0; j<nW; j++ )  {

                niX = (j-nW2) * nScaleMult;
                niY = (i-nH2) * nScaleMult;

                nIter = 0;
                while( nIter ^ 1<<nIterBits )
                {    
                    niXX = niX>>ONEBITS2;  niXX *= niXX;
                    niYY = niY>>ONEBITS2;  niYY *= niYY;

                    if( niXX+niYY > 4<<ONEBITS )
                        break;

                    niY = (niX>>ONEBITS2)*(niY>>ONEBITS21) + niCY;
                    niX = niXX - niYY                      + niCX;

                    nIter++;
                }

                BYTE col = 255 - (nIter << (7-nIterBits));

               *pLine++ = col;
               *pLine++ = col;
               *pLine++ = col;
            }
        }
        
    }

}


void BlitBg( HDC hdc )
{
    HDC     dc = CreateCompatibleDC( hdc );
    HGDIOBJ ob = SelectObject( dc, hbBg );

    if( bRedraw )
    {
        RedrawBg();

        HDC     dcBG = CreateCompatibleDC( hdc );
        HGDIOBJ obBG = SelectObject( dcBG, hbBgDib );
        BitBlt( dc, 0,0,nW,nH, dcBG, 0,0, SRCCOPY );
        SelectObject( dcBG, obBG );
        DeleteDC( dcBG );

        bRedraw = 0;
    }

    BitBlt( hdc, 0,0,nW,nH, dc, 0,0, SRCCOPY );
    SelectObject( dc, ob );
    DeleteDC( dc );
}




void DrawPolyline( HDC hdc )
{
    //----------- Det
    int nIters[MAXPOLYLINE][2];


    long nW2 = nW>>1;
    long nH2 = nH>>1;

    long nScaleMult = long( (1 << ONEBITS-7) / pow( 2, nScale/8.0 ) );  // 8 raz dlja scale=x2
                                                                        // 7 - default scale
    long niX = (nMouseShiftX-nW2) * nScaleMult;
    long niY = (nMouseShiftY-nH2) * nScaleMult;

    long niCX = (nMouseX-nW2) * nScaleMult;
    long niCY = (nMouseY-nH2) * nScaleMult;


    int nIter = 0;
    while( nIter ^ 1<<nPolylineBits  &&  nIter < MAXPOLYLINE )
    {
        nIters[nIter][0] = niX/nScaleMult + nW2;
        nIters[nIter][1] = niY/nScaleMult + nH2;
        nIter++;

        long niXX = niX>>ONEBITS2;  niXX *= niXX;
        long niYY = niY>>ONEBITS2;  niYY *= niYY;

        if( niXX+niYY > 12<<ONEBITS )
            break;

        niY = (niX>>ONEBITS2)*(niY>>ONEBITS21) + niCY;
        niX = niXX - niYY                      + niCX;

    }


    //----------- Draw
    int i = nIter;

    SelectObject( hdc, hPens[ MAXPEN * i/nIter ] );
    MoveToEx( hdc, nIters[i-1][0],
                   nIters[i-1][1], NULL );

    for( ; i; i-- )  {

        SelectObject( hdc, hPens[ MAXPEN * i/nIter ] );
        LineTo( hdc, nIters[i-1][0],
                     nIters[i-1][1] );
    }


}


void DrawImage( HDC hdc )
{
    HBRUSH  hbg = (HBRUSH)GetStockObject( NULL_BRUSH );
    HGDIOBJ hob = SelectObject( hdc, hbg );
    HGDIOBJ hop = SelectObject( hdc, hPens[0xCF] );

     BlitBg( hdc );
     if( bPolylineShow )
        DrawPolyline( hdc );

     if( bHelpShow )  {
         RECT rect;
         SetRect( &rect, 10,10,300,300 );
         SetBkMode( hdc, TRANSPARENT );
         DrawText( hdc, szAbout, strlen(szAbout), &rect, DT_EXPANDTABS );
     }


    SelectObject( hdc, hop );
    SelectObject( hdc, hob );
}


//========================================== Screen

void CreateScreen( HDC hdc )  {

    int nScrH = GetSystemMetrics( SM_CYSCREEN );
    int nScrW = GetSystemMetrics( SM_CXSCREEN );


    // Mandelbrot Buffer
    pBgDibBitsMand = (BYTE*)GlobalAlloc( GMEM_FIXED, nScrW*nScrH );


    // Mem Bitmaps
    hbScreen = CreateCompatibleBitmap( hdc, nScrW,nScrH );

    hbBg     = CreateCompatibleBitmap( hdc, nScrW,nScrH );

    // DIB
    BITMAPINFO bi = {0};
    bi.bmiHeader.biSize   = sizeof( BITMAPINFOHEADER );
    bi.bmiHeader.biWidth  = nScrW;
    bi.bmiHeader.biHeight = nScrH;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount    = 24;
    bi.bmiHeader.biCompression = BI_RGB; 

    hbBgDib = CreateDIBSection( hdc, &bi, DIB_RGB_COLORS, (LPVOID*)&pBgDibBits, 0,0 );

    BITMAP bm;
    GetObject( hbBgDib, sizeof(BITMAP), &bm );
    nBgDibWidthBytes = bm.bmWidthBytes;
    nBgDibHeight     = nScrH;


    // Pens
    for( int i=0; i<=MAXPEN; i++ )
      hPens[i] = CreatePen( PS_SOLID, 1, 0x010101 * i );

}
void DeleteScreen()  {
    for( int i=0; i<=MAXPEN; i++ )
      DeleteObject( hPens[i] );
    DeleteObject( hbBgDib  );
    DeleteObject( hbBg     );
    DeleteObject( hbScreen );

    if( pBgDibBits )  GlobalFree( pBgDibBits );
}


void DrawScreen( HWND hWnd, HDC hdc )  {
    RECT rect;
    SetRect( &rect,0,0,nW,nH );

    HDC dcScreen = CreateCompatibleDC( hdc );
     HGDIOBJ ob = SelectObject( dcScreen, hbScreen );
      
      FillRect ( dcScreen, &rect, (HBRUSH)(COLOR_WINDOW+1) );
      DrawImage( dcScreen );

      BitBlt( hdc, 0,0,nW,nH, dcScreen, 0,0, SRCCOPY );
     SelectObject( dcScreen, ob );
    DeleteDC( dcScreen );
}

//========================================== Window

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC         hdc;

	switch( message )  {

        case WM_SIZE:
            if( SIZE_MAXIMIZED == wParam || 
                SIZE_RESTORED  == wParam )  {
                nW = LOWORD(lParam);
                nH = HIWORD(lParam);
                bRedraw = REDRAW_ALL;
                InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
            }


		case WM_PAINT:
			hdc = BeginPaint( hWnd, &ps );
             if( !hbScreen )  CreateScreen( hdc );     // Init Once
             DrawScreen( hWnd, hdc );
            EndPaint( hWnd, &ps );
			break;


        case WM_SYSCOMMAND:
            if( IDM_ABOUT == (wParam & 0xFFF0) )  {     // About
                MessageBox( hWnd, szAbout, "About", MB_OK );
                return 0;
            }
            else
                return DefWindowProc( hWnd, message, wParam, lParam );


		case WM_KEYDOWN:
            switch( wParam )  {

                case 27:    // Esc
                    DestroyWindow( hWnd );
                    break;

                case 112:   // F1   About
                    bHelpShow ^= 1;
                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    //MessageBox( hWnd, szAbout, "About", MB_OK );
                    break;

                case 116:   // F5   Redraw
                    bRedraw = REDRAW_ALL;
                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    break;

                case 122:   // F11  Toggle FullScreen
                    bFullScreen ^= 1;

                    if( bFullScreen )  {
                        SetWindowLong( hWnd, GWL_STYLE, WS_POPUP );
                        SetWindowPos ( hWnd, 0, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_FRAMECHANGED );
                        ShowWindow( hWnd, SW_MAXIMIZE );
                    }
                    else  {
                        ShowWindow( hWnd, SW_RESTORE );
                        SetWindowLong( hWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW );
                        SetWindowPos ( hWnd, 0, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_FRAMECHANGED | SWP_SHOWWINDOW );
                    }
                    break;                    

                case 107:   // +
                case 187:
                case 109:   // -
                case 189:
                    if( wParam&3 ^ 3 )
                    {   if( nIterBits > 1 )  nIterBits--;  }
                    else
                    {   if( nIterBits < 7 )  nIterBits++;  }

                    bRedraw = REDRAW_ALL;
                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    break;


                case 80: // P   Toggle Polyline Show
                    bPolylineShow ^= 1;
                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    break;

                case 77: // M   Toggle Mandelbrot Show
                    bMandShow ^= 1;
                    bRedraw = REDRAW_ALL;
                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    break;

                case 37: //  <
                case 39: //  >
                    if( wParam & 2 )
                    {   if( nPolylineBits < 9 )  nPolylineBits++;  }
                    else
                        if( nPolylineBits > 1 )  nPolylineBits--;

                    InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
                    break;


                default:
                    return DefWindowProc( hWnd, WM_KEYDOWN, wParam, lParam );
            }
            break;

        case WM_MOUSEWHEEL: // Zoom
            if( (short)HIWORD(wParam) > 0 )  nScale ++;
            else                             nScale --;

            bRedraw = REDRAW_ALL;
            InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
            break;

        case WM_MOUSEMOVE:
            nMouseShiftX = LOWORD(lParam);
            nMouseShiftY = HIWORD(lParam);
            if(!(MK_SHIFT & wParam  ||  MK_LBUTTON & wParam ) )  {
                nMouseX = nMouseShiftX;
                nMouseY = nMouseShiftY;
			    bRedraw |= REDRAW_JULIA;
            }

            InvalidateRect( hWnd, NULL, FALSE );    // Redraw Window
            break;

		case WM_DESTROY:
			PostQuitMessage(0);
            DeleteScreen();
			break;

		default:
			return DefWindowProc( hWnd, message, wParam, lParam );
    }
    return 0;
}





BOOL InitWindow( HINSTANCE hInstance, int nCmdShow )
{
    //---------------------------------- Register Class
	WNDCLASSEX wcex = {0};
	wcex.cbSize         = sizeof(WNDCLASSEX); 
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.hInstance		= hInstance;
	wcex.hCursor		= LoadCursor( NULL,IDC_ARROW );
	wcex.hbrBackground	= (HBRUSH)COLOR_WINDOWTEXT;
	wcex.lpszClassName	= szTitle;

	RegisterClassEx( &wcex );


    //---------------------------------- Create Window
    HWND hWnd = CreateWindow( szTitle, szCaption, WS_OVERLAPPEDWINDOW,
                              CW_USEDEFAULT,CW_USEDEFAULT,
                              CW_USEDEFAULT,CW_USEDEFAULT,
                              NULL,NULL, hInstance,NULL );
    if( !hWnd )
        return FALSE;


    // Add About to System menu
    HMENU hSysMenu = GetSystemMenu( hWnd, FALSE );
	if( hSysMenu )	{
        AppendMenu( hSysMenu, MF_SEPARATOR, 0,0 );
        AppendMenu( hSysMenu, MF_STRING, IDM_ABOUT, "About\tF1" );
	}



    ShowWindow( hWnd, nCmdShow );
    UpdateWindow( hWnd );

    return TRUE;
}



int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int nCmdShow) 
{
	if( !InitWindow( hInstance, nCmdShow ) ) 
		return 0;

	MSG msg;
	while( GetMessage( &msg,NULL,0,0 ) )  {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}



